@@ -6,11 +6,12 @@
 ############# Auxiliary non-exported functions #############
 
 OPTS_PDN <- c("positive", "negative", "default")
-HYPS_NAMES <- c("thetas", "shapes", "scales", "alphas", "gammas")
+HYPS_NAMES <- c("thetas", "shapes", "scales", "alphas", "gammas", "penalty")
 HYPS_OTHERS <- c("lambda", "train_size")
 LEGACY_PARAMS <- c("cores", "iterations", "trials", "intercept_sign", "nevergrad_algo")
 
-check_nas <- function(df) {
+check_nas <- function(df, channels = NULL) {
+  if (!is.null(channels)) df <- select(df, all_of(channels))
   name <- deparse(substitute(df))
   if (sum(is.na(df)) > 0) {
     naVals <- lares::missingness(df)
@@ -38,16 +39,24 @@ check_novar <- function(dt_input, InputCollect = NULL) {
       "There are %s column(s) with no-variance: %s. \nPlease, remove the variable(s) to proceed...",
       length(novar), v2t(novar)
     )
-    if (!is.null(InputCollect)) msg <- sprintf(
-      "%s\n>>> Note: there's no variance on these variables because of the modeling window filter (%s:%s)",
-      msg,
-      InputCollect$window_start,
-      InputCollect$window_end
-    )
+    if (!is.null(InputCollect)) {
+      msg <- sprintf(
+        "%s\n>>> Note: there's no variance on these variables because of the modeling window filter (%s:%s)",
+        msg,
+        InputCollect$window_start,
+        InputCollect$window_end
+      )
+    }
     stop(msg)
   }
 }
 
+check_allneg <- function(df) {
+  all_negative <- unlist(lapply(df, function(x) all(x <= 0)))
+  df <- mutate_at(df, names(which(all_negative)), function(x) abs(x))
+  return(df)
+}
+
 check_varnames <- function(dt_input, dt_holidays,
                            dep_var, date_var,
                            context_vars, paid_media_spends,
@@ -163,24 +172,38 @@ check_prophet <- function(dt_holidays, prophet_country, prophet_vars, prophet_si
   } else {
     prophet_vars <- tolower(prophet_vars)
     opts <- c("trend", "season", "monthly", "weekday", "holiday")
+    if (!"holiday" %in% prophet_vars) {
+      if (!is.null(prophet_country)) {
+        warning(paste(
+          "Input 'prophet_country' is defined as", prophet_country,
+          "but 'holiday' is not setup within 'prophet_vars' parameter"
+        ))
+      }
+      prophet_country <- NULL
+    }
     if (!all(prophet_vars %in% opts)) {
       stop("Allowed values for 'prophet_vars' are: ", paste(opts, collapse = ", "))
     }
     if ("weekday" %in% prophet_vars && dayInterval > 7) {
       warning("Ignoring prophet_vars = 'weekday' input given your data granularity")
     }
-    if (is.null(prophet_country) || length(prophet_country) > 1 |
-      !prophet_country %in% unique(dt_holidays$country)) {
+    if ("holiday" %in% prophet_vars && (
+      is.null(prophet_country) || length(prophet_country) > 1 |
+        isTRUE(!prophet_country %in% unique(dt_holidays$country)))) {
       stop(paste(
         "You must provide 1 country code in 'prophet_country' input.",
         length(unique(dt_holidays$country)), "countries are included:",
         paste(unique(dt_holidays$country), collapse = ", "),
-        "\nIf your country is not available, please manually add it to 'dt_holidays'"
+        "\nIf your country is not available, manually include data to 'dt_holidays'",
+        "or remove 'holidays' from 'prophet_vars' input."
       ))
     }
     if (is.null(prophet_signs)) {
       prophet_signs <- rep("default", length(prophet_vars))
     }
+    if (length(prophet_signs) == 1) {
+      prophet_signs <- rep(prophet_signs, length(prophet_vars))
+    }
     if (!all(prophet_signs %in% OPTS_PDN)) {
       stop("Allowed values for 'prophet_signs' are: ", paste(OPTS_PDN, collapse = ", "))
     }
@@ -224,7 +247,7 @@ check_paidmedia <- function(dt_input, paid_media_vars, paid_media_signs, paid_me
   check_vector(paid_media_vars)
   check_vector(paid_media_signs)
   check_vector(paid_media_spends)
-  mediaVarCount <- length(paid_media_vars)
+  expVarCount <- length(paid_media_vars)
   spendVarCount <- length(paid_media_spends)
 
   temp <- paid_media_vars %in% names(dt_input)
@@ -242,7 +265,7 @@ check_paidmedia <- function(dt_input, paid_media_vars, paid_media_signs, paid_me
     ))
   }
   if (is.null(paid_media_signs)) {
-    paid_media_signs <- rep("positive", mediaVarCount)
+    paid_media_signs <- rep("positive", expVarCount)
   }
   if (!all(paid_media_signs %in% OPTS_PDN)) {
     stop("Allowed values for 'paid_media_signs' are: ", paste(OPTS_PDN, collapse = ", "))
@@ -253,13 +276,17 @@ check_paidmedia <- function(dt_input, paid_media_vars, paid_media_signs, paid_me
   if (length(paid_media_signs) != length(paid_media_vars)) {
     stop("Input 'paid_media_signs' must have same length as 'paid_media_vars'")
   }
-  if (spendVarCount != mediaVarCount) {
+  if (spendVarCount != expVarCount) {
     stop("Input 'paid_media_spends' must have same length as 'paid_media_vars'")
   }
   is_num <- unlist(lapply(dt_input[, paid_media_vars], is.numeric))
   if (!all(is_num)) {
     stop("All your 'paid_media_vars' must be numeric. Check: ", v2t(paid_media_vars[!is_num]))
   }
+  is_num <- unlist(lapply(dt_input[, paid_media_spends], is.numeric))
+  if (!all(is_num)) {
+    stop("All your 'paid_media_spends' must be numeric. Check: ", v2t(paid_media_spends[!is_num]))
+  }
   get_cols <- any(dt_input[, unique(c(paid_media_vars, paid_media_spends))] < 0)
   if (get_cols) {
     check_media_names <- unique(c(paid_media_vars, paid_media_spends))
@@ -272,7 +299,7 @@ check_paidmedia <- function(dt_input, paid_media_vars, paid_media_signs, paid_me
   }
   return(invisible(list(
     paid_media_signs = paid_media_signs,
-    mediaVarCount = mediaVarCount,
+    expVarCount = expVarCount,
     paid_media_vars = paid_media_vars
   )))
 }
@@ -300,14 +327,17 @@ check_organicvars <- function(dt_input, organic_vars, organic_signs) {
   if (length(organic_signs) != length(organic_vars)) {
     stop("Input 'organic_signs' must have same length as 'organic_vars'")
   }
+  is_num <- unlist(lapply(dt_input[, organic_vars], is.numeric))
+  if (!all(is_num)) {
+    stop("All your 'organic_vars' must be numeric. Check: ", v2t(organic_vars[!is_num]))
+  }
   return(invisible(list(organic_signs = organic_signs)))
 }
 
-check_factorvars <- function(dt_input, factor_vars = NULL, context_vars = NULL, organic_vars = NULL) {
+check_factorvars <- function(dt_input, factor_vars = NULL, context_vars = NULL) {
   check_vector(factor_vars)
   check_vector(context_vars)
-  check_vector(organic_vars)
-  temp <- select(dt_input, all_of(c(context_vars, organic_vars)))
+  temp <- select(dt_input, all_of(context_vars))
   are_not_numeric <- !sapply(temp, is.numeric)
   if (any(are_not_numeric)) {
     these <- are_not_numeric[!names(are_not_numeric) %in% factor_vars]
@@ -318,8 +348,8 @@ check_factorvars <- function(dt_input, factor_vars = NULL, context_vars = NULL,
     }
   }
   if (!is.null(factor_vars)) {
-    if (!all(factor_vars %in% c(context_vars, organic_vars))) {
-      stop("Input 'factor_vars' must be any from 'context_vars' or 'organic_vars' inputs")
+    if (!all(factor_vars %in% context_vars)) {
+      stop("Input 'factor_vars' must be any from 'context_vars' inputs")
     }
   }
   return(factor_vars)
@@ -440,7 +470,8 @@ check_adstock <- function(adstock) {
 
 check_hyperparameters <- function(hyperparameters = NULL, adstock = NULL,
                                   paid_media_spends = NULL, organic_vars = NULL,
-                                  exposure_vars = NULL) {
+                                  exposure_vars = NULL, prophet_vars = NULL,
+                                  contextual_vars = NULL) {
   if (is.null(hyperparameters)) {
     message(paste(
       "Input 'hyperparameters' not provided yet. To include them, run",
@@ -465,6 +496,15 @@ check_hyperparameters <- function(hyperparameters = NULL, adstock = NULL,
     ref_all_media <- sort(c(ref_hyp_name_spend, ref_hyp_name_org, HYPS_OTHERS))
     all_ref_names <- c(ref_hyp_name_spend, ref_hyp_name_expo, ref_hyp_name_org, HYPS_OTHERS)
     all_ref_names <- all_ref_names[order(all_ref_names)]
+    # Adding penalty variations to the dictionary
+    if (any(grepl("_penalty", paste0(get_hyp_names)))) {
+      ref_hyp_name_penalties <- paste0(
+        c(paid_media_spends, organic_vars, prophet_vars, contextual_vars), "_penalty"
+      )
+      all_ref_names <- c(all_ref_names, ref_hyp_name_penalties)
+    } else {
+      ref_hyp_name_penalties <- NULL
+    }
     if (!all(get_hyp_names %in% all_ref_names)) {
       wrong_hyp_names <- get_hyp_names[which(!(get_hyp_names %in% all_ref_names))]
       stop(
@@ -473,7 +513,7 @@ check_hyperparameters <- function(hyperparameters = NULL, adstock = NULL,
       )
     }
     total <- length(get_hyp_names)
-    total_in <- length(c(ref_hyp_name_spend, ref_hyp_name_org, ref_hyp_name_other))
+    total_in <- length(c(ref_hyp_name_spend, ref_hyp_name_org, ref_hyp_name_penalties, ref_hyp_name_other))
     if (total != total_in) {
       stop(sprintf(
         paste(
@@ -641,6 +681,26 @@ check_calibration <- function(dt_input, date_var, calibration_input, dayInterval
   return(calibration_input)
 }
 
+check_obj_weight <- function(calibration_input, objective_weights, refresh) {
+  obj_len <- ifelse(is.null(calibration_input), 2, 3)
+  if (!is.null(objective_weights)) {
+    if ((length(objective_weights) != obj_len)) {
+      stop(paste0("objective_weights must have length of ", obj_len))
+    }
+    if (any(objective_weights < 0) | any(objective_weights > 10)) {
+      stop("objective_weights must be >= 0 & <= 10")
+    }
+  }
+  if (is.null(objective_weights) & refresh) {
+    if (obj_len == 2) {
+      objective_weights <- c(0, 1)
+    } else {
+      objective_weights <- c(0, 1, 1)
+    }
+  }
+  return(objective_weights)
+}
+
 check_iteration <- function(calibration_input, iterations, trials, hyps_fixed, refresh) {
   if (!refresh) {
     if (!hyps_fixed) {
@@ -660,7 +720,7 @@ check_InputCollect <- function(list) {
   names_list <- c(
     "dt_input", "paid_media_vars", "paid_media_spends", "context_vars",
     "organic_vars", "all_ind_vars", "date_var", "dep_var",
-    "rollingWindowStartWhich", "rollingWindowEndWhich", "mediaVarCount",
+    "rollingWindowStartWhich", "rollingWindowEndWhich",
     "factor_vars", "prophet_vars", "prophet_signs", "prophet_country",
     "intervalType", "dt_holidays"
   )
@@ -777,11 +837,7 @@ check_init_msg <- function(InputCollect, cores) {
   if (cores == 1) {
     message(paste(base, "with no parallel computation"))
   } else {
-    if (check_parallel()) {
-      message(paste(base, "on", cores, "cores"))
-    } else {
-      message(paste(base, "on 1 core (Windows fallback)"))
-    }
+    message(paste(base, "on", cores, "cores"))
   }
 }
 
@@ -790,6 +846,9 @@ check_class <- function(x, object) {
 }
 
 check_allocator_constrains <- function(low, upr) {
+  if (all(is.na(low)) || all(is.na(upr))) {
+    stop("You must define lower (channel_constr_low) and upper (channel_constr_up) constraints")
+  }
   max_length <- max(c(length(low), length(upr)))
   if (any(low < 0)) {
     stop("Inputs 'channel_constr_low' must be >= 0")
@@ -804,7 +863,6 @@ check_allocator_constrains <- function(low, upr) {
 
 check_allocator <- function(OutputCollect, select_model, paid_media_spends, scenario,
                             channel_constr_low, channel_constr_up, constr_mode) {
-  check_allocator_constrains(channel_constr_low, channel_constr_up)
   if (!(select_model %in% OutputCollect$allSolutions)) {
     stop(
       "Provided 'select_model' is not within the best results. Try any of: ",
@@ -816,6 +874,7 @@ check_allocator <- function(OutputCollect, select_model, paid_media_spends, scen
   if (!(scenario %in% opts)) {
     stop("Input 'scenario' must be one of: ", paste(opts, collapse = ", "))
   }
+  check_allocator_constrains(channel_constr_low, channel_constr_up)
   if (!(scenario == "target_efficiency" & is.null(channel_constr_low) & is.null(channel_constr_up))) {
     if (length(channel_constr_low) != 1 && length(channel_constr_low) != length(paid_media_spends)) {
       stop(paste(
@@ -861,15 +920,16 @@ check_metric_dates <- function(date_range = NULL, all_dates, dayInterval = NULL,
   ## default using latest 30 days / 4 weeks / 1 month for spend level
   if (is.null(date_range)) {
     if (is.null(dayInterval)) stop("Input 'date_range' or 'dayInterval' must be defined")
-    if (!is_allocator) {
-      date_range <- "last_1"
-    } else {
-      date_range <- paste0("last_", case_when(
-        dayInterval == 1 ~ 30,
-        dayInterval == 7 ~ 4,
-        dayInterval >= 30 & dayInterval <= 31 ~ 1,
-      ))
-    }
+    # if (!is_allocator) {
+    #   date_range <- "last_1"
+    # } else {
+    #   date_range <- paste0("last_", case_when(
+    #     dayInterval == 1 ~ 30,
+    #     dayInterval == 7 ~ 4,
+    #     dayInterval >= 30 & dayInterval <= 31 ~ 1,
+    #   ))
+    # }
+    date_range <- "all"
     if (!quiet) message(sprintf("Automatically picked date_range = '%s'", date_range))
   }
   if (grepl("last|all", date_range[1])) {
